#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <math.h>

#include "adaptiveAttribute.h"
#include "JVector.h"
#include "attributesInfo.h"
#include "discretizationManager.h"
#include "timerADI.h"
#include "timerGlobals.h"
#include "random.h"
#include "instanceSet.h"

extern attributesInfo ai;
extern discretizationManager disc;
extern timerADI *tADI;
extern timerGlobals *tGlobals;
extern Random rnd;
extern instanceSet *is;

#define min(a,b) (((a)<(b))?(a):(b))

void adaptiveAttributeConstructor(int *attribute, int whichAttribute)
{
	int i;

	if (tADI->typeOfAttribute[whichAttribute] == REAL) {
		// Discretization algorithm assigned to this attribute
		int num = disc.getNumDiscretizers();
		attribute[1] = rnd(0, num - 1);

		// Number of microIntervasl we have to distribute among all 
		// intervals
		int microIntervalsOfAttribute =
		    disc.getDiscretizer(attribute[1])
		    ->getNumIntervals(whichAttribute);

		// Number of intervals of the attribute
		int numIntervals;
		int maxAllowedIntervals =
		    min(tADI->numMaxIntervals, microIntervalsOfAttribute);
		if (maxAllowedIntervals > 2)
			numIntervals = rnd(2, maxAllowedIntervals);
		else
			numIntervals = maxAllowedIntervals;

		attribute[0] = numIntervals;

		// Size of each interval
		int sum=0;
		for (i = 0; i < numIntervals - 1; i++) {
			int numMicroIntervals = rnd(1,
						    microIntervalsOfAttribute
						    - (numIntervals - 1 -
						       i));
			sum+=numMicroIntervals;
			attribute[i * 2 + 2] = sum;
			microIntervalsOfAttribute -= numMicroIntervals;
		}
		sum+=microIntervalsOfAttribute;
		attribute[(numIntervals - 1) * 2 + 2] = sum;
	} else {
		attribute[0] = attribute[1] =
		    ai.getNumValuesAttribute(whichAttribute);
	}
}

void adaptiveAttributeInitializeIntervalValues(int *attribute,
					       int whichAttribute,
					       instance * ins)
{
	int i;

	if (tADI->typeOfAttribute[whichAttribute] == REAL) {
		int intervalOfSample;
		int done = 0;

		if (ins) {
			intervalOfSample = ins->nominalValuesFromDiscretizers[attribute[1]*tGlobals->numAttributesMC+whichAttribute];
		} else {
			done = 1;
		}

		for (i = 0; i < attribute[0]; i++) {
			int value;
			if (!rnd < tGlobals->probOne) {
				value = 1;
			} else {
				value = 0;
			}
			attribute[i * 2 + 3] = value;

			if (!done) {
				if (intervalOfSample < attribute[i * 2 + 2]) {
					done = 1;
					attribute[i * 2 + 3] = 1;
				}
			}
		}
	} else {
		int valueIns;
		if (ins)
			valueIns = ins->valueOfAttribute(whichAttribute);
		else
			valueIns = -1;
		for (i = 0; i < attribute[0]; i++) {
			int value;
			if (i != valueIns) {
				if (!rnd < tGlobals->probOne) {
					value = 1;
				} else {
					value = 0;
				}
			} else {
				value = 1;
			}
			attribute[i + 2] = value;
		}
	}
}

void adaptiveAttributeDoReinitialize(int *attribute, int whichAttribute)
{
	if (tADI->typeOfAttribute[whichAttribute] != REAL)
		return;
	tADI->numReinitialize++;
	adaptiveAttributeConstructor(attribute, whichAttribute);
	adaptiveAttributeInitializeIntervalValues(attribute,
						  whichAttribute, NULL);
}

void adaptiveAttributeDoMerge(int *attribute, int whichAttribute,
			      int interval)
{
	int i;

	tADI->numMergeAttempts++;

	if (tADI->typeOfAttribute[whichAttribute] != REAL
	    || attribute[0] == 1)
		return;

	tADI->numMerge++;


	//We choose the neighbour we are going to merge with 
	int neighbour;
	if (interval == 0) {
		neighbour = 1;
	} else if (interval == attribute[0] - 1) {
		neighbour = attribute[0] - 2;
	} else {
		if (!rnd < 0.5) {
			neighbour = interval - 1;
		} else {
			neighbour = interval + 1;
		}
	}

	//Sort them
	int pos1, pos2, nextInterval;
	if (neighbour < interval) {
		pos1 = neighbour * 2 + 2;
		nextInterval = interval + 1;
	} else {
		pos1 = interval * 2 + 2;
		nextInterval = neighbour + 1;
	}
	pos2 = pos1 + 2;

	int size1, size2;
	if(pos1==2) {
		size1=attribute[pos1];
	} else {
		size1=attribute[pos1]-attribute[pos1-2];
	}
	size2=attribute[pos2]-attribute[pos1];



	// Value of the merged interval 
	int value;
	if (size1 > size2) {
		value = attribute[pos1 + 1];
	} else if (size2 > size1) {
		value = attribute[pos2 + 1];
	} else {
		if (!rnd < 0.5) {
			value = attribute[pos1 + 1];
		} else {
			value = attribute[pos2 + 1];
		}
	}

	// Set the merged interval
	attribute[pos1] = attribute[pos2];
	attribute[pos1 + 1] = value;

	// Set all the intervals after the new one
	for (i = nextInterval; i < attribute[0]; i++, pos2 += 2) {
		attribute[pos2] = attribute[pos2 + 2];
		attribute[pos2 + 1] = attribute[pos2 + 3];
	}

	attribute[0]--;
}


void adaptiveAttributeDoSplit(int *attribute, int whichAttribute,
			      int interval)
{
	int i;

	if (tADI->typeOfAttribute[whichAttribute] != REAL)
		return;

	//If interval has only one microinterval or the maximum number of
	//intervals has been reached, we can't apply the operation
	int intervalPos = interval * 2 + 2;
	int size;
	if(intervalPos==2) {
		size=attribute[intervalPos];
	} else {
		size=attribute[intervalPos]-attribute[intervalPos-2];
	}
	if (size < 2 || attribute[0] == tADI->numMaxIntervals) return;
	tADI->numSplit++;

	//We select the cut point. We cut after this point
	int cutPoint = rnd(0, size - 2);
	int size1 = cutPoint + 1;
	int size2 = size - size1;

	//We have to move one forward position all intervals after the splitted
	//one
	int pos = (attribute[0] - 1) * 2 + 2;
	for (i = attribute[0] - 1; i >= interval; i--, pos -= 2) {
		attribute[pos + 2] = attribute[pos];
		attribute[pos + 3] = attribute[pos + 1];
	}

	attribute[intervalPos]-=size2;

	attribute[0]++;
}

void adaptiveAttributeMutation(int *attribute, int whichAttribute)
{
#ifdef DEBUG
	if (interval < 0 || interval >= attribute[0]) {
		fprintf(stderr, "Out of bound interval at "
			"adaptiveAttribute::Mutacio:%d\n", interval);
		exit(1);
	}
#endif

	int interval = rnd(0, attribute[0] - 1);
	int pos;
	if (tADI->typeOfAttribute[whichAttribute] == REAL) {
		pos = interval * 2 + 3;
	} else {
		pos = interval + 2;
	}

	if (attribute[pos] == 1) {
		attribute[pos] = 0;
	} else {
		attribute[pos] = 1;
	}
}

void adaptiveAttributeDumpPhenotype(int *attribute, int whichAttribute,
				    char *string)
{
	int i, j;
	int pos = 0;
	char temp[50000];
	char temp2[1000];

	strcpy(string, "");
	sprintf(temp, "Att %s is ",
		ai.getAttributeName(whichAttribute)->cstr());
	if (tADI->typeOfAttribute[whichAttribute] == REAL) {
		double min = ai.getMinDomain(whichAttribute);
		double max = ai.getMaxDomain(whichAttribute);
		discretizer *d = disc.getDiscretizer(attribute[1]);

		if (attribute[0] > 1) {
			int attrCount = 0;
			int index = 0;
			int oldMicroInt = 0;
			while (index < attribute[0]) {
				attrCount++;
				int microInt;
				int valueAct = attribute[index * 2 + 3];
				while (index < attribute[0]
				       && attribute[index * 2 + 3] ==
				       valueAct) {
					microInt=attribute[index * 2 + 2];
					index++;
				}
				if (valueAct) {
					if (oldMicroInt == 0) {
						if(index<attribute[0]) {
							double maxInt = d->getCutPoint(whichAttribute
								,microInt - 1);
							sprintf(temp2,"[<%g]",maxInt);
						} else {
							strcpy(temp2,"");
						}
					} else {
						double minInt = d->getCutPoint(whichAttribute
							,oldMicroInt - 1);
						if (index == attribute[0]) {
							sprintf(temp2,"[>%g]",minInt);
						} else {
							double maxInt = d->getCutPoint(whichAttribute
								,microInt - 1);
							sprintf(temp2,"[%g,%g]",minInt,maxInt);
						}
					}
					strcat(temp, temp2);
				}
				oldMicroInt = microInt;
			}

			if (attrCount > 1) {
				strcpy(string, temp);
			}
		}
	} else {
		int irr = 1;
		for (i = 0; i < attribute[0]; i++) {
			if (attribute[i + 2]) {
				sprintf(temp2, "%s,",
					ai.getNominalValue(whichAttribute,
							   i)->cstr());
				strcat(temp, temp2);
			} else {
				irr = 0;
			}
		}

		if (!irr) {
			if (temp[strlen(temp) - 1] == ',')
				temp[strlen(temp) - 1] = 0;
			strcpy(string, temp);
		}
	}
}

void adaptiveAttributeDisable(int *attribute,int whichAttribute) 
{
	int i,index;
	if (tADI->typeOfAttribute[whichAttribute] == REAL) {
		for(i=0,index=3;i<attribute[0];i++,index+=2) {
			attribute[index]=0;
		}
	} else {
		for(i=0,index=2;i<attribute[0];i++,index++) {
			attribute[index]=0;
		}
	}
}

void adaptiveAttributeSetValue(int *attribute,int whichAttribute,int value)
{
	int i,index;

	if (tADI->typeOfAttribute[whichAttribute] == REAL) {
		int interval;
		for(i=0,index=2;i<attribute[0];i++,index+=2) {
			if(value<attribute[index]) {
				interval=i;
				break;
			}
		}
		attribute[interval*2+3]=1;
	} else {
		attribute[value+2]=1;
	}
}

