#include "instanceSetVirtual.h"
#include "configManagement.h"
#include "instance.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "JVector.h"
#include "JString.h"
#include "attributesInfo.h"
#include "utils.h"
#include "windowingVirtualSampling.h"
#include "random.h"
#include "messageBuffer.h"
#include "timeManagement.h"
#include "mxGenerator.h"


extern messageBuffer mb;
extern attributesInfo ai;
extern Random rnd;
extern configManagement cm;

void instanceSetVirtual::generateDataset(int traintest)
{
	unsigned long long i;

	if(cm.thereIsParameter(DATASET_MX)) {
		gen = new mxGenerator((int)cm.getParameter(DATASET_MX));
	} else {
		fprintf(stderr,"Dataset Not Defined !!!\n");
		exit(1);
	}

	numInstances=(unsigned long long)gen->getNumInstances();

	if(traintest==TRAIN) {
		gen->initialize();
		/*temp=new instance();
		for(i=0;i<numInstances;i++) {
			if(!(i%1000000)) {
				printf("Foo %u\n",i);
			}
			gen->generateInstance(temp,i);	
			ai.insertInstance(temp);
		}*/

		ai.calculateAverages();
	}
}

void instanceSetVirtual::initializeWindowing(int traintest)
{
	if(traintest == TRAIN) {
		if(cm.thereIsParameter(WINDOWING_ILAS)) {
			windowingEnabled=1;
			win=new windowingVirtualSampling;
		} else {
			windowingEnabled=0;
			window=NULL;
		}

		if(windowingEnabled) {
			win->setInstances(gen);
			win->newIteration(window,windowSize);
		}
	} else {
		windowingEnabled=0;
		window=NULL;
	}
}

instanceSetVirtual::instanceSetVirtual(int traintest)
{
	int i;

	generateDataset(traintest);
	initializeWindowing(traintest);

	if(traintest==TRAIN) {
		numClasses=ai.getNumClasses();
		classWiseInit=cm.thereIsParameter(CLASS_WISE_INIT);
		initInstanceLists();
	} else {
		initSamplings=NULL;
		countsByClass=NULL;
		instByClass=NULL;
	}
}

instanceSetVirtual::~instanceSetVirtual()
{
	int i;

	if(initSamplings) {
		for(i=0;i<numClasses;i++) {
			delete initSamplings[i];
			delete instByClass[i];
		}
		delete initSamplings;
		delete instByClass;
		delete countsByClass;
	}
	//if(window) delete window;
}

unsigned long long instanceSetVirtual::getNumInstancesOfIteration()
{
	if(windowingEnabled) return windowSize;
	return numInstances;
}

int instanceSetVirtual::newIteration(int isLast)
{
	/*if(isLast) {
		//mb.printf("Windowing disabled for last iteration\n");
		windowingEnabled=0;
		return 1;
	}*/

	if(windowingEnabled) {
		win->newIteration(window,windowSize);
		if(win->needReEval()) return 1;
		return 0;
	}

	return 0;
}

void instanceSetVirtual::initInstanceLists()
{
	int i;

	countsByClass = new int[numClasses];
	initSamplings = new Sampling *[numClasses];
	instByClass = new int *[numClasses];

	if(windowingEnabled) {
		int numInst=(int)getNumInstancesOfIteration();

		for(i=0;i<numInst;i++) {
			int cl=window[i]->instanceClass;
			countsByClass[cl]++;
		}

		for(i=0;i<numClasses;i++) {
			instByClass[i] = new int[countsByClass[i]];
			countsByClass[i]=0;
		}
	
		for(i=0;i<numInst;i++) {
			int cl=window[i]->instanceClass;
			instByClass[cl][countsByClass[cl]++]=i;
		}

		for(i=0;i<numClasses;i++) {
			initSamplings[i] = new Sampling(countsByClass[i]);
		}
	} else {
		fprintf(stderr,"Instance-based initialization without windowing is not supported !!!\n");
		exit(1);
	}
}
	
instance *instanceSetVirtual::getInstanceInit(int forbiddenCL) 
{
	if(classWiseInit) {
		int cl;
		do {
			cl=rnd(0,numClasses-1);
		} while(cl==forbiddenCL);

		int pos=initSamplings[cl]->getSample();
		int insIndex=instByClass[cl][pos];
		instance *ins;

		if(windowingEnabled) ins=window[insIndex];
		else { 
			fprintf(stderr,"Instance-based initialization without windowing is not supported !!!\n");
			exit(1);
		}

		return ins;
	} else {
		int nc=numClasses;
		int count[nc];
		int total=0;
		int i;
		if(forbiddenCL!=numClasses) nc--;
	
		for(i=0;i<nc;i++) {
			if(i<forbiddenCL) 
				count[i]=initSamplings[i]->numSamplesLeft();
			else 
				count[i]=initSamplings[i+1]->numSamplesLeft();
			total+=count[i];
		}
		int pos=rnd(0,total-1);
		int acum=0;
		int found=0;
		for(i=0;i<nc && !found;i++) {
			acum+=count[i];
			if(pos<acum) {
				found=1;
			}
		}
		i--;
		if(i>=forbiddenCL) i++;

		pos=initSamplings[i]->getSample();
		int insIndex=instByClass[i][pos];

		instance *ins;
		if(windowingEnabled) ins=window[insIndex];
		else { 
			fprintf(stderr,"Instance-based initialization without windowing is not supported !!!\n");
			exit(1);
		}

		return ins;
	}
}

int instanceSetVirtual::numVersions()
{
        if(windowingEnabled) return win->numVersions();
        return 1;
}

int instanceSetVirtual::getCurrentVersion()
{
        if(windowingEnabled) return win->getCurrentVersion();
        return 0;
}

