#include "mpi.h"
#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <sys/times.h>
#include <unistd.h>

#include "configManagement.h"
#include "populationWrapper.h"
#include "classifierFitness.h"
#include "random.h"
#include "instanceSetFile.h"
#include "timersManagement.h"
#include "discretizationManager.h"
#include "attributesInfo.h"
#include "lex_conf.h"
#include "timeManagement.h"
#include "messageBuffer.h"
#include "timerCrossover.h"
#include "timerEvolutionStats.h"

int stop = 0;
messageBuffer mb;
discretizationManager disc;
attributesInfo ai;
configManagement cm;
instanceSet *is;
timeManagement timeM;
Random rnd;
double percentageOfLearning=0;
int lastIteration=0;
int nodeRank;
int useMPI=1;
int numTasks;
extern timerCrossover *tCross;
extern timerEvolutionStats *tEvolStats;

void handler(int sig)
{
	stop = 1;
}

void doCrossover(classifierFactory *cf)
{
	JVector<classifier *>individuals;
	int flag,i;

	MPI_Recv(&flag,1,MPI_CHAR,0,0,MPI_COMM_WORLD,MPI_STATUS_IGNORE);
	while(flag) {
		int numParents=tCross->numParentsSmartX;
		classifier *parents[numParents];
		for(i=0;i<numParents;i++) {
			parents[i]=cf->createClassifierMPI();
		}
		classifier *son=cf->cloneClassifier(parents[0],1);
		son->crossover_smart(parents,numParents);
		individuals.addElement(son);
		for(i=0;i<numParents;i++) {
			cf->deleteClassifier(parents[i]);
		}

		MPI_Recv(&flag,1,MPI_CHAR,0,0,MPI_COMM_WORLD,MPI_STATUS_IGNORE);
	}

	int num=individuals.size();
	MPI_Request reqs[2*num];
	for(i=0;i<num;i++) {
		individuals[i]->sendIndividual(0,&reqs[i*2]);
		cf->deleteClassifier(individuals[i]);
	}
	MPI_Waitall(2*num,reqs,MPI_STATUSES_IGNORE);
}

void doLS(classifierFactory *cf)
{
	JVector<classifier *>individuals;
	int flag,i;

	MPI_Recv(&flag,1,MPI_CHAR,0,0,MPI_COMM_WORLD,MPI_STATUS_IGNORE);
	while(flag) {
		classifier *ind=cf->createClassifierMPI();
		ind->doLocalSearch();
		individuals.addElement(ind);
		MPI_Recv(&flag,1,MPI_CHAR,0,0,MPI_COMM_WORLD,MPI_STATUS_IGNORE);
	}

	int num=individuals.size();
	MPI_Request reqs[2*num];
	for(i=0;i<num;i++) {
		individuals[i]->sendIndividual(0,&reqs[i*2]);
		cf->deleteClassifier(individuals[i]);
	}
	MPI_Waitall(2*num,reqs,MPI_STATUSES_IGNORE);
}



int main(int argc, char *argv[])
{
	int rc;

	if (argc < 4) {
		fprintf(stderr,"Incorrect parameters\n"
		       "%s: <Config file> <Train file> <Test file>\n",
		       argv[0]);
		exit(1);
	}


	rc = MPI_Init(&argc,&argv);
        if (rc != MPI_SUCCESS) {
                printf ("Error starting MPI program. Terminating.\n");
                MPI_Abort(MPI_COMM_WORLD, rc);
        }
        MPI_Comm_size(MPI_COMM_WORLD,&numTasks);
        MPI_Comm_rank(MPI_COMM_WORLD,&nodeRank);

	if(nodeRank==0 && argc==5) mb.setFile(argv[4]);

	//printf("I am %d of %d\n",nodeRank,numTasks);

	if(numTasks==1) useMPI=0;

	if(nodeRank!=0) mb.ignoreMessages();
	parseConfig(argv[1]);
	if(nodeRank!=0) mb.allowMessages();

	if(useMPI) {
		MPI_Barrier(MPI_COMM_WORLD);
		if(nodeRank==0) {
			unsigned long int seed=rnd.getSeed();
			int ret=MPI_Bcast(&seed,1,MPI_UNSIGNED_LONG,0,MPI_COMM_WORLD);
		} else {
			unsigned long int seed;
			int ret=MPI_Bcast(&seed,1,MPI_UNSIGNED_LONG,0,MPI_COMM_WORLD);
			rnd.setSeed(seed);
		}
	}

	if(nodeRank==0) {
		rnd.dumpSeed();
	}

	if(nodeRank!=0) mb.ignoreMessages();

	is=new instanceSetFile(argv[2], TRAIN);
	timersManagement timers;
	timeM.resetTime();

	if(useMPI) {
		MPI_Barrier(MPI_COMM_WORLD);
	}

	if(nodeRank==0) {
		populationWrapper pw((int)cm.getParameter(POP_SIZE));

		timers.setPW(&pw);
		signal(SIGINT, handler);

		double ave,dev;
		pw.getAverageDevAccuracy(ave,dev);
		mb.printf("Initial population average accuracy:%f deviation:%f\n",ave,dev);
		double totalTime=-1;
		if(cm.thereIsParameter(TOTAL_TIME)) 
			totalTime=cm.getParameter(TOTAL_TIME);
	
		int countIt=0,numIterations = (int) cm.getParameter(ITERATIONS);
		if(numIterations==0) stop=1;
	
		int dumpGenotype=-1;
		if(cm.thereIsParameter(DUMP_GENOTYPE_ITERATIONS)) {
			dumpGenotype=(int)cm.getParameter(DUMP_GENOTYPE_ITERATIONS);
		}

	
		for (;!stop;) {
			timers.incIteration(lastIteration);
	
			if (is->newIteration(lastIteration)) {
				pw.activateModifiedFlag();
			}
	
			if(dumpGenotype!=-1 && !(countIt%dumpGenotype)) {
				classifier *ind=pw.getBestPopulation();
				char string[100000];
				mb.printf("Dumping genotype\n");
				ind->dumpGenotype(string);
				mb.printf("%s\n",string);
			}
				
				
				

			pw.gaIteration();

			classifier *ind=pw.getBestOverall();
			//char pheno[1000000];
			//ind->dumpPhenotype(pheno);
			//mb.printf("Phenotype: \n%s\n", pheno);
			if(classifierTest(*ind,is)==1.0) {
				stop=1;
			}

			if(useMPI) {
				MPI_Bcast(&stop,1,MPI_UNSIGNED_LONG,0,MPI_COMM_WORLD);
			}
	
			timers.dumpStats();
			countIt++;
			if(totalTime==-1) {
				if(countIt==numIterations) stop=1;
				else if(countIt==numIterations-1) lastIteration=1;
				percentageOfLearning=(double)countIt
					/(double)(numIterations);
			} else {
				if(lastIteration) stop=1;
				else {
					double currentTime=timeM.totalTime();
					percentageOfLearning=currentTime/totalTime;
					if(percentageOfLearning>1) 
						percentageOfLearning=1;
					if(currentTime>=totalTime) {
						mb.printf("Total number of iterations in time %d : %d\n",(int)totalTime,countIt+1);
						lastIteration=1;
					}
				}
			}
			mb.printf("Total time: %g\n",timeM.totalTime());
		}
	
		classifier *ind = pw.getBestOverall();
	
		//ind->postProcessIndividual();
	
		char phenotype[1000000];
		ind->dumpPhenotype(phenotype);
		mb.printf("Phenotype: \n%s\n", phenotype);
	
	
		classifierStats(*ind, argv[2], "Train");
		classifierStats(*ind, argv[3], "Test");
	} else {
		classifierFactory *cf= new classifierFactory;

		classifier *ind=cf->createClassifier();
		int hasSmartX=ind->hasSmartCrossover() && (tCross->smartCrossoverProb>0);
		int hasLS=ind->hasLocalSearch() && (tGlobals->probLocalSearch>0);
		cf->deleteClassifier(ind);

		evaluateIndividuals(cf);

		int countIt=0,numIterations = (int) cm.getParameter(ITERATIONS);
		if(numIterations==0) stop=1;
	
		for (;!stop;) {
			is->newIteration(lastIteration);

			if(hasSmartX && !lastIteration) doCrossover(cf);
			if(hasLS && !lastIteration) doLS(cf);
	
			evaluateIndividuals(cf);
			MPI_Bcast(&stop,1,MPI_UNSIGNED_LONG,0,MPI_COMM_WORLD);

			MPI_Send(&tEvolStats->numExampleEvaluations, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);

			countIt++;
			if(countIt==numIterations) stop=1;
			else if(countIt==numIterations-1) lastIteration=1;
		}

		delete cf;
	}

	delete is;
	
	if(useMPI) {
		MPI_Barrier(MPI_COMM_WORLD);
	}

	MPI_Finalize();

	return 0;
}
